import pytest
from ..util import dictionary_lookup, DictionaryKeyError

def test_dictionary_lookup():

    d1 = {'a': 3, 'b': 4}
    assert dictionary_lookup(d1, 'a') == 3
    assert dictionary_lookup(d1, 'b') == 4

    # TODO: These next few would be easier to do in unittest.TestCase where we could use self.assertRaisesRegexpe.
    #       Maybe a later version of pytest has the equivalent feature. -kmp 9-Feb-2020

    d2 = {'x': 10}
    try:
        dictionary_lookup(d2, 'y')
    except Exception as e:
        assert isinstance(e, DictionaryKeyError)
        assert str(e) == '''{'x': 10} has no 'y' key.'''
    else:
        raise AssertionError("No exception was raised where one was expected.")

    try:
        dictionary_lookup(17, 'z')
    except Exception as e:
        assert isinstance(e, DictionaryKeyError)
        assert str(e) == '''17 is not a dictionary.'''
    else:
        raise AssertionError("No exception was raised where one was expected.")

    try:
        # String form of JSON isn't what's needed. It has to be parsed (i.e., a dict).
        dictionary_lookup(repr(d2), 'x')
    except Exception as e:
        assert isinstance(e, DictionaryKeyError)
        assert str(e) == '''"{'x': 10}" is not a dictionary.'''
    else:
        raise AssertionError("No exception was raised where one was expected.")



"""
Example:

    %(prog)s *.tsv

"""

from .. import loadxl
import json
import os.path

EPILOG = __doc__


def rename_test_with_underscore(rows):
    for row in rows:
        if 'test' in row:
            if row['test'] != 'test':
                row['_test'] = row['test']
            del row['test']
        yield row


def remove_empty(rows):
    for row in rows:
        if row:
            yield row


def convert(filename, sheetname=None, outputdir=None, skip_blanks=False):
    if outputdir is None:
        outputdir = os.path.dirname(filename)
    source = loadxl.read_single_sheet(filename, sheetname)
    pipeline = [
        loadxl.remove_keys_with_empty_value if skip_blanks else loadxl.noop,
        rename_test_with_underscore,
        remove_empty,
    ]
    data = list(loadxl.combine(source, pipeline))
    if sheetname is None:
        sheetname, ext = os.path.splitext(os.path.basename(filename))
    out = open(os.path.join(outputdir, sheetname + '.json'), 'w')
    json.dump(data, out, sort_keys=True, indent=4, separators=(',', ': '))



def main():
    import argparse
    parser = argparse.ArgumentParser(
        description="Convert spreadsheet to json list", epilog=EPILOG,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser.add_argument('filenames', metavar='FILE', nargs='+', help="Files to convert")
    parser.add_argument('--outputdir', help="Directory to write converted output")
    parser.add_argument('--sheetname', help="xlsx sheet name")
    parser.add_argument('--skip-blanks', help="Skip blank columns")
    args = parser.parse_args()

    for filename in args.filenames:
        convert(filename, args.sheetname, args.outputdir, args.skip_blanks)


if __name__ == '__main__':
    main()



from urllib.parse import quote
import os.path
import sys
try:
    import subprocess32 as subprocess
except ImportError:
    import subprocess


def initdb(datadir, prefix='', echo=False):
    init_args = [
        os.path.join(prefix, 'initdb'),
        '-D', datadir,
        '-U', 'postgres',
        '--auth=trust',
    ]
    output = subprocess.check_output(
        init_args,
        close_fds=True,
        stderr=subprocess.STDOUT,
    )
    if echo:
        print(output.decode('utf-8'))


def server_process(datadir, prefix='', echo=False):
    args = [
        os.path.join(prefix, 'postgres'),
        '-D', datadir,
        '-F',  # no fsync
        '-h', '',
        '-k', datadir,
        '-p', '5432'
    ]
    process = subprocess.Popen(
        args,
        close_fds=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
    )

    SUCCESS_LINE = b'database system is ready to accept connections\n'

    lines = []
    for line in iter(process.stdout.readline, b''):
        if echo:
            sys.stdout.write(line.decode('utf-8'))
        lines.append(line)
        if line.endswith(SUCCESS_LINE):
            break
    else:
        code = process.wait()
        msg = ('Process return code: %d\n' % code) + b''.join(lines).decode('utf-8')
        raise Exception(msg)

    if not echo:
        process.stdout.close()

    if echo:
        print('Created: postgresql://postgres@:5432/postgres?host=%s' % quote(datadir))

    return process


def main():
    import atexit
    import shutil
    import tempfile
    datadir = tempfile.mkdtemp()

    print('Starting in dir: %s' % datadir)
    try:
        process = server_process(datadir, echo=True)
    except:
        shutil.rmtree(datadir)
        print('Cleaned dir: %s' % datadir)
        raise

    @atexit.register
    def cleanup_process():
        try:
            if process.poll() is None:
                process.terminate()
                for line in process.stdout:
                    sys.stdout.write(line.decode('utf-8'))
                process.wait()
        finally:
            shutil.rmtree(datadir)
            print('Cleaned dir: %s' % datadir)

    print('Started. ^C to exit.')

    try:
        for line in iter(process.stdout.readline, b''):
            sys.stdout.write(line.decode('utf-8'))
    except KeyboardInterrupt:
        raise SystemExit(0)


if __name__ == '__main__':
    main()



import sys
if sys.version_info.major == 2:
    from future.standard_library import install_aliases
    install_aliases()
    import functools
    from backports.functools_lru_cache import lru_cache
    functools.lru_cache = lru_cache
import netaddr
from pyramid.config import Configurator
from pyramid.settings import (
    asbool,
)
from .calculated import calculated_property  # noqa
from .config import (  # noqa
    abstract_collection,
    collection,
    root,
)
from .interfaces import *  # noqa
from .resources import (  # noqa
    AbstractCollection,
    Collection,
    Item,
    Resource,
    Root,
    display_title_schema
)
from .schema_utils import load_schema  # noqa
from .upgrader import upgrade_step  # noqa
from .app import (
    app_version,
    session,
    configure_dbsession,
    changelogs,
    json_from_path,
    )
import logging
import os
from dcicutils.log_utils import set_logging
from dcicutils.ff_utils import get_health_page


def includeme(config):
    config.include('pyramid_retry')
    config.include('pyramid_tm')
    config.include('.util')
    config.include('.stats')
    config.include('.batchupgrade')
    config.include('.calculated')
    config.include('.config')
    config.include('.connection')
    config.include('.embed')
    config.include('.json_renderer')
    config.include('.validation')
    config.include('.predicates')
    config.include('.invalidation')
    config.include('.upgrader')
    config.include('.aggregated_items')
    config.include('.storage')
    config.include('.typeinfo')
    config.include('.resources')
    config.include('.attachment')
    config.include('.schema_graph')
    config.include('.jsonld_context')
    config.include('.schema_views')
    config.include('.crud_views')
    config.include('.indexing_views')
    config.include('.resource_views')


def main(global_config, **local_config):
    """ This function returns a Pyramid WSGI application.
    """
    settings = global_config
    settings.update(local_config)

    # adjust log levels for some annoying loggers
    lnames = ['boto', 'urllib', 'elasticsearch', 'dcicutils']
    for name in logging.Logger.manager.loggerDict:
        if any(logname in name for logname in lnames):
            logging.getLogger(name).setLevel(logging.WARNING)

    set_logging(in_prod=settings.get('production'))
    # set_logging(settings.get('elasticsearch.server'), settings.get('production'))

    # TODO - these need to be set for dummy app
    # settings['snovault.jsonld.namespaces'] = json_asset('snovault:schemas/namespaces.json')
    # settings['snovault.jsonld.terms_namespace'] = 'https://www.encodeproject.org/terms/'
    settings['snovault.jsonld.terms_prefix'] = 'snovault'

    config = Configurator(settings=settings)
    from .elasticsearch import APP_FACTORY
    config.registry[APP_FACTORY] = main  # used by mp_indexer
    config.include(app_version)

    config.include('pyramid_multiauth')  # must be before calling set_authorization_policy
    from pyramid_localroles import LocalRolesAuthorizationPolicy
    # Override default authz policy set by pyramid_multiauth
    config.set_authorization_policy(LocalRolesAuthorizationPolicy())
    config.include(session)

    config.include(configure_dbsession)
    config.include('snovault')
    config.commit()  # commit so search can override listing

    config.include('.renderers')

    # only include this stuff if we're testing
    if asbool(settings.get('testing', False)):
        config.include('snovault.tests.testing_views')
        config.include('snovault.tests.authentication')
        config.include('snovault.tests.root')
        if settings.get('elasticsearch.server'):
            config.include('snovault.tests.search')

    if 'elasticsearch.server' in config.registry.settings:
        config.include('snovault.elasticsearch')

    config.include(changelogs)

    # TODO This is optional AWS only - possibly move to a plug-in
    aws_ip_ranges = json_from_path(settings.get('aws_ip_ranges_path'), {'prefixes': []})
    config.registry['aws_ipset'] = netaddr.IPSet(
        record['ip_prefix'] for record in aws_ip_ranges['prefixes'] if record['service'] == 'AMAZON')

    # cache mirror_health in registry if need be
    mirror = settings.get('mirror.env.name', None)
    if mirror:
        settings['mirror_health'] = get_health_page(ff_env=mirror)

    app = config.make_wsgi_app()

    return app



import json
import pytest
import structlog
import time
import yaml

from unittest import mock
from .. import stats
from ..crud_views import log as crud_view_log
from .test_post_put_patch import COLLECTION_URL, item_with_uuid


@pytest.fixture
def mocked():
    return mock.patch.object(stats, 'log')


def test_stats_tween_logs_stats(testapp, mocked):
    ''' plus in this case we always log url '''
    with mocked as mocked_log :
        testapp.get("/")
        assert mocked_log.bind.call_count == 2
        assert mocked_log.bind.call_args_list[0] == mock.call(url_path='/',
                                                              url_qs='',
                                                              host='localhost')
        mocked_log.bind.assert_called_with(db_count=mock.ANY, db_time=mock.ANY,
                                                rss_begin=mock.ANY, rss_change=mock.ANY,
                                                rss_end=mock.ANY, wsgi_begin=mock.ANY,
                                                url_path='/',
                                                url_qs='',
                                                host='localhost',
                                                wsgi_end=mock.ANY, wsgi_time=mock.ANY)
    return


def test_stats_tween_logs_telemetry_id(testapp, mocked):
    with mocked as mocked_log:
        res = testapp.get("/?telemetry_id=test_telem")
        assert mocked_log.bind.call_count == 2
        assert mocked_log.bind.call_args_list[0] == mock.call(telemetry_id='test_telem',
                                                              url_path='/',
                                                              url_qs='telemetry_id=test_telem',
                                                              host='localhost')
        mocked_log.bind.assert_called_with(db_count=mock.ANY, db_time=mock.ANY,
                                           rss_begin=mock.ANY, rss_change=mock.ANY,
                                           rss_end=mock.ANY, wsgi_begin=mock.ANY,
                                           wsgi_end=mock.ANY, wsgi_time=mock.ANY,
                                           url_path='/',
                                           url_qs='telemetry_id=test_telem',
                                           host='localhost',
                                           telemetry_id='test_telem')

        # we should also return telem in the header
        assert 'telemetry_id=test_telem' in res.headers['X-Stats']
    return


def test_telemetry_id_carries_through_logging(testapp, external_tx):
    mocked = mock.patch.object(crud_view_log, 'info')
    with mocked as mock_log:
        res = testapp.post_json(COLLECTION_URL + "?telemetry_id=test&log_action=action_test", item_with_uuid[0], status=201)
        mock_log.assert_called_with(event="add_to_indexing_queue", uuid=mock.ANY,
                                    sid=mock.ANY, telemetry_id=mock.ANY)
        # also make sure we have a logger that has defaultsset from stats.py
        logger = crud_view_log.bind()
        assert logger._context.get('url_path') == COLLECTION_URL
        assert logger._context.get('url_qs') == "telemetry_id=test&log_action=action_test"
        assert logger._context.get('host') == 'localhost'
        assert logger._context.get('telemetry_id') == 'test'
        assert logger._context.get('log_action') == 'action_test'


def test_logging_basic(testapp, external_tx, capfd):
        '''
        in prod logging setup, an Elasticsearch server is provided. Logs will
        be piped to the appropriate logs (e.g. httpd/error_log) and also sent
        to Elasticsearch. That is tested here in snovault in test_indexing;
        here, we configure the logs without the es_server to ensure that
        the rest of it works
        '''
        # something that generates logs
        # add a telemetry id and some log contents using a query string
        res = testapp.post_json(COLLECTION_URL + "?telemetry_id=test&log_action=action_test", item_with_uuid[0], status=201)
        # multiple logs emitted in this process, must find the one we want
        check_logs = capfd.readouterr()[-1].split('\n')
        log_msg = None
        for record in check_logs:
            if not record:
                continue
            try:
                proc_record = yaml.safe_load('{' + record.strip().split('{', 1)[1])
            except:
                continue
            if not isinstance(proc_record, dict):
                continue
            if proc_record.get('telemetry_id') == 'test':
                log_msg = proc_record
        assert '@timestamp' in log_msg
        assert 'logger' in log_msg
        assert 'level' in log_msg


def test_logging_see_debug_log(testapp, capfd):
    """
    Tests that when we hit a route with the @debug_log decorator we see an appropriate log statement
    """
    testapp.get('/')  # all routes are marked
    check_logs = capfd.readouterr()[-1].split('\n')
    for record in check_logs:
        if not record:
            continue
        if 'DEBUG_FUNC' in record:
            return
    raise AssertionError("Did not see 'DEBUG_FUNC' in a log message")



import sys
from itertools import islice
from urllib.parse import urlencode

from future.utils import (
    raise_with_traceback,
    itervalues,
)
from past.builtins import basestring
from pyramid.exceptions import PredicateMismatch
from pyramid.httpexceptions import HTTPNotFound
from pyramid.settings import asbool
from pyramid.view import (
    render_view_to_response,
    view_config,
)

from .calculated import calculate_properties
from .resources import (
    AbstractCollection,
    Item,
    Root,
)
from .util import (
    expand_path,
    build_embedded_model,
    expand_embedded_model,
    process_aggregated_items,
    check_es_and_cache_linked_sids,
    validate_es_content,
    debug_log,
)
from .util import filter_embedded


def includeme(config):
    config.scan(__name__)


@view_config(context=AbstractCollection, permission='list', request_method='GET',
             name='listing')
@debug_log
def collection_view_listing_db(context, request):
    result = {}

    frame = request.params.get('frame', 'columns')

    limit = request.params.get('limit', 25)
    if limit in ('', 'all'):
        limit = None
    if limit is not None:
        try:
            limit = int(limit)
        except ValueError:
            limit = 25

    items = (
        item for item in itervalues(context)
        if request.has_permission('view', item)
    )

    if limit is not None:
        items = islice(items, limit)

    result['@graph'] = [
        request.embed(request.resource_path(item, '@@' + frame))
        for item in items
    ]

    if limit is not None and len(result['@graph']) == limit:
        params = [(k, v) for k, v in request.params.items() if k != 'limit']
        params.append(('limit', 'all'))
        result['all'] = '%s?%s' % (request.resource_path(context), urlencode(params))

    return result


@view_config(context=Root, request_method='GET', name='page')
@debug_log
def home(context, request):
    properties = request.embed(request.resource_path(context), '@@object')
    calculated = calculate_properties(context, request, properties, category='page')
    properties.update(calculated)
    return properties


@view_config(context=Root, request_method='GET', name='object')
@view_config(context=AbstractCollection, permission='list', request_method='GET', name='object')
@debug_log
def collection_view_object(context, request):
    properties = context.__json__(request)
    calculated = calculate_properties(context, request, properties)
    properties.update(calculated)
    return properties


@view_config(context=AbstractCollection, permission='list', request_method='GET', name='page')
@debug_log
def collection_list(context, request):
    path = request.resource_path(context)
    properties = request.embed(path, '@@object')
    calculated = calculate_properties(context, request, properties, category='page')
    properties.update(calculated)

    if request.query_string:
        properties['@id'] += '?' + request.query_string

    result = request.embed(path, '@@listing?' + request.query_string, as_user=True)
    result.update(properties)
    return result


@view_config(context=Root, request_method='GET')
@view_config(context=AbstractCollection, permission='list', request_method='GET')
@view_config(context=Item, permission='view', request_method='GET')
@debug_log
def item_view(context, request):
    frame = request.params.get('frame', 'page')
    if getattr(request, '__parent__', None) is None:
        # We need the response headers from non subrequests
        try:
            response = render_view_to_response(context, request, name=frame)
        except PredicateMismatch:
            # Avoid this view emitting PredicateMismatch
            exc_class, exc, tb = sys.exc_info()
            exc.__class__ = HTTPNotFound
            raise_with_traceback(exc, tb)
        else:
            if response is None:
                raise HTTPNotFound('?frame=' + frame)
            return response
    path = request.resource_path(context, '@@' + frame)
    if request.query_string:
        path += '?' + request.query_string
    return request.embed(path, as_user=True)


@view_config(context=Item, permission='view', request_method='GET',
             name='object')
@debug_log
def item_view_object(context, request):
    """
    Render json structure. This is the most basic view and contains the scope
    of only one item. The steps to generate it are:
    1. Fetch stored properties, possibly upgrading.
    2. Link canonicalization (overwriting uuids with links)
       - adds uuid to request._linked_uuids if request._indexing_view
    3. Calculated properties

    On a DB request, will use the Elasticsearch result for the view if the ES
    result passes `validate_es_content` (has valid sids and rev_links)

    Args:
        context: current Item
        request: current Request

    Returns:
        Dictionary item properties
    """
    if request.datastore == 'elasticsearch':
        es_res = check_es_and_cache_linked_sids(context, request, 'object')
        # validate_es_content also checks/updates rev links
        if es_res and validate_es_content(context, request, es_res, 'object'):
            # if indexing, handle linked_uuids
            if request._indexing_view is True:
                request._linked_uuids = [link['uuid'] for link in es_res['linked_uuids_object']]
            return es_res['object']

    properties = context.item_with_links(request)
    calculated = calculate_properties(context, request, properties)
    properties.update(calculated)
    return properties


@view_config(context=Item, permission='view', request_method='GET', name='embedded')
@debug_log
def item_view_embedded(context, request):
    """
    Calculate and return the embedded view for an item. This is an intensive
    process that requires an embedded model to be created from the
    context.embedded_list and then build from traversal of individual object
    views from the items needed.
    This function also takes care of processing aggregated_items, which are
    held on the request object (and transferred to subrequests in embed.py)

    On a DB request, will use the Elasticsearch result for the view if the ES
    result passes `validate_es_content` (has valid sids and rev_links)

    Args:
        context: current Item
        request: current Request

    Returns:
        Dictionary item properties
    """
    if request.datastore == 'elasticsearch':
        es_res = check_es_and_cache_linked_sids(context, request, 'embedded')
        # validate_es_content also checks/updates rev links
        if es_res and validate_es_content(context, request, es_res, 'embedded'):
            # if indexing, handle aggregated_items and linked_uuids
            if request._indexing_view is True:
                request._linked_uuids = [link['uuid'] for link in es_res['linked_uuids_embedded']]
            if getattr(request, '_aggregate_for').get('uuid') == str(context.uuid):
                # format this in a specific way to work with further processing
                request._aggregated_items = {agg: {'items': val} for agg, val in
                                             es_res['aggregated_items'].items()}
                request._aggregate_for['uuid'] = None
            return filter_embedded(es_res['embedded'], request.effective_principals)

    # set up _aggregated_items if we want to aggregate this target
    will_aggregate = getattr(request, '_aggregate_for').get('uuid') == str(context.uuid)
    if will_aggregate:
        parent_path = request.resource_path(context)
    else:
        parent_path = None

    item_path = request.resource_path(context)
    properties = request.embed(item_path, '@@object', as_user=True)
    embedded_model = build_embedded_model(context.embedded)
    embedded = expand_embedded_model(request, properties, embedded_model,
                                     parent_path=parent_path)
    if will_aggregate:
        process_aggregated_items(request)
        # we don't want to aggregate this item again, such as if the same item
        # is used in a frame=embedded calc property.
        # use a list here so that the reference is maintained through subreq
        request._aggregate_for['uuid'] = None

    return embedded


@view_config(context=Item, permission='view', request_method='GET',
             name='page')
@debug_log
def item_view_page(context, request):
    item_path = request.resource_path(context)
    properties = request.embed(item_path, '@@embedded', as_user=True)
    calculated = calculate_properties(context, request, properties, category='page')
    properties.update(calculated)
    return properties


@view_config(context=Item, permission='expand', request_method='GET',
             name='expand')
@debug_log
def item_view_expand(context, request):
    path = request.resource_path(context)
    properties = request.embed(path, '@@object', as_user=True)
    for path in request.params.getall('expand'):
        expand_path(request, properties, path)
    return properties


def expand_column(request, obj, subset, path):
    if isinstance(path, basestring):
        path = path.split('.')
    if not path:
        return
    name = path[0]
    remaining = path[1:]
    value = obj.get(name, None)
    if value is None:
        return
    if not remaining:
        subset[name] = value
        return
    if isinstance(value, list):
        if name not in subset:
            subset[name] = [{} for i in range(len(value))]
        for index, member in enumerate(value):
            if not isinstance(member, dict):
                member = request.embed(member, '@@object')
            expand_column(request, member, subset[name][index], remaining)
    else:
        if name not in subset:
            subset[name] = {}
        if not isinstance(value, dict):
            value = request.embed(value, '@@object')
        expand_column(request, value, subset[name], remaining)


@view_config(context=Item, permission='view', request_method='GET',
             name='columns')
@debug_log
def item_view_columns(context, request):
    path = request.resource_path(context)
    properties = request.embed(path, '@@object')
    if context.schema is None or 'columns' not in context.schema:
        return properties

    subset = {
        '@id': properties['@id'],
        '@type': properties['@type'],
    }

    for column in context.schema['columns']:
        path = column.split('.')
        if path[-1] == 'length':
            path.pop()
        if path:
            expand_column(request, properties, subset, path)

    return subset


@view_config(context=Item, permission='view_raw', request_method='GET',
             name='raw')
@debug_log
def item_view_raw(context, request):
    props = context.properties
    # only upgrade properties if explicitly requested
    if asbool(request.params.get('upgrade', True)):
        props =  context.upgrade_properties()
    # add uuid to raw view
    props['uuid'] = str(context.uuid)
    return props
